unit TestCurveFitForm;

{
   Main Form for TestCurveFit Delphi Program 

Purpose: to provide a simple visual test app for the procedure PolyFit

Revision history:

2000 Oct 28  First version
2001 Jan 07  Add UpDown button to demonstrate changing the number of terms
             Add more explanatory comments
             Build without requiring Delphi 4 runtime library
2003 Mar 05  Rebuild on Delphi 5
             Add check for errors option when all Y values are the same

2007 Mar 28  Changed ancient pascal 'Real' types to 'Double'.

Copyright  David J Taylor, Edinburgh and others listed above
Web site:  www.satsignal.net           
E-mail:    davidtaylor@writeme.com
}

interface

// Add unit CurveFit to the Uses clause to gain access to PolyFit
uses
  Windows, Messages, SysUtils, Classes, Controls, Forms,
  Dialogs, StdCtrls, ComCtrls, Graphics, Math,
  CurveFit, ExtCtrls;

const
  num_points = 9;
  max_terms = 5;
  // plot constants:
  diamondSize=5; // pixel size of the Diamond marker.

type
  TFormMain = class (TForm)
    MemoResults: TMemo;
    ButtonClose: TButton;
    UpDownNumTerms: TUpDown;
    EditNumTerms: TEdit;
    LabelNumTerms: TLabel;
    paintMe: TPaintBox;
    CheckBoxOriginForce: TCheckBox;
    procedure FormShow (Sender: TObject);
    procedure ButtonCloseClick (Sender: TObject);
    procedure UpDownNumTermsClick (Sender: TObject; Button: TUDBtnType);
    procedure paintMePaint(Sender: TObject);
    procedure CheckBoxOriginForceClick(Sender: TObject);
  private
    FValid : Boolean;

    { Private declarations }
    procedure calculate_fit (const nTerms: integer);

    procedure PlotDiamond(X, Y: Integer);
    procedure PlotFilledDiamond(X, Y: Integer);

  public

    nPoints: Integer;
    x, y: Array of Double; // was Array of Real
    coef: Array of Double; // was Array of Real
    plotStart,plotEnd:Integer; // Range to plot onto chart
    correl: Double;
    { Public declarations }
  end;

var
  FormMain: TFormMain;

implementation

{$R *.DFM}


procedure TFormMain.FormShow (Sender: TObject);
begin
  // On first show, calcuate some results to display...
  Calculate_fit (UpDownNumTerms.Position);
end;


procedure TFormMain.ButtonCloseClick (Sender: TObject);
begin
  // When the Close button is clicked, close the form and application
  Close;
end;


procedure TFormMain.UpDownNumTermsClick(Sender: TObject; Button: TUDBtnType);
begin
  // When the UpDown control is clicked, try
  // a curve fit with a new number of terms
  Calculate_fit (UpDownNumTerms.Position);
end;


procedure TFormMain.calculate_fit (const nTerms: integer);
var
  i, j: integer;
  xc, yc, delta: real;
begin
  plotStart := 0;
  plotEnd := 0;

  FValid := false;
  // Sanity check on arguments
  if (nTerms < 1) or (nTerms > max_terms) then Exit;

  // Fill in the X and Y arrays
  nPoints := 9;

  SetLength(x,nPoints);
  SetLength(y,nPoints);
  SetLength(coef, nTerms+1);


  if (nTerms>2) then begin // show something we can fit a curve to
      for i := 0 to nPoints - 1 do
          x [i] := i + 1;

    y [0] := 2.07;
    y [1] := 8.6;
    y [2] := 14.42;
    y [3] := 15.80;
    y [4] := 18.92;
    y [5] := 17.96;
    y [6] := 12.98;
    y [7] := 6.45;
    y [8] := 0.27;
  end else begin // show something we can fit a line to
      for i := 0 to nPoints - 1 do
          x [i] := i + 1;
  
    y [0] := 1.67;
    y [1] := 2.37;
    y [2] := 4.82;
    y [3] := 6.81;
    y [4] := 6.99;
    y [5] := 5.99;
    y [6] := 7.84;
    y [7] := 9.84;
    y [8] := 16.32;

    x[8] := 18;

  end;
  
  //------------------------------------------------------------
  //
  // To force a point through the origin, you must provide
  // a doubled set of sample data, the duplicates have the
  // opposite sign, so if your original data is
  // [ [1,1.1], [2,2.2], [3,3.3] ]
  // then feed in
  // [ [-3,-3.3], [-2,-2.2], [-1,-1.1], [1,1.1], [2,2.2], [3,3.3] ]
  //
  // and you will get a curve that goes through the origin.
  //
  // Note that the correlation coefficient is weaker in
  // significance, in this case, since you have added "imaginary"
  // negative data, your corelation coefficient will GO UP
  // which would indicate a better fit, but in this case, the
  // fit is WORSE because of being forced through the origin.
  // Don't be confused by that. Look at the sum of the squares
  // of the residuals, which gets higher, indicating a poorer fit
  // of the sample data.  However, in some applications this
  // 'poorer' fit is actually the fit you want, because you know
  // that 0 of something really must correspond to 0 of something
  // else, and so you constrain your regression to go through
  // the origin.
  //------------------------------------------------------------
  if  CheckBoxOriginForce.Checked then begin
      SetLength(x,nPoints*2);
      SetLength(y,nPoints*2);
      // copy forward into second half of array.
      for i := 0 to nPoints-1 do begin
          x[nPoints+i] := x[i];
          y[nPoints+i] := y[i];
      end;
      // make a copy with with opposite signs: (negative to positive, vice versa)
      for i := 1 to nPoints do begin
          x[i-1] := -1*x[nPoints+(nPoints-i)];
          y[i-1] := -1*y[nPoints+(nPoints-i)];
      end;

      PolyFit (x, y, coef, correl, nPoints*2, nTerms);
      plotStart := nPoints;
      plotEnd := (nPoints*2)-1; // plot original values, not duplicates.
  end else begin
      { NORMAL: }
      PolyFit (x, y, coef, correl, nPoints, nTerms);
      plotStart := 0;
      plotEnd := nPoints - 1; // plot everything.
  end;

  // Remove any existing display of results
  MemoResults.Clear;


  // Display the results in the lines of a MemoControl
  MemoResults.Lines.Add ('     X      Y    Ycalc  residual');


  for i := plotStart to plotEnd do
    begin
    yc := 0.0;
    xc := 1.0;
    for j := 0 to nTerms - 1 do
      begin
      yc := yc + coef [j] * xc;
      xc := xc * x [i];
      end;
    delta := yc - y [i];
    MemoResults.Lines.Add (Format ('   %3f %6.2f %6.2f %6.2f',
      [ x [i], y [i], yc, delta]));
    end;
  MemoResults.Lines.Add (' ');
  MemoResults.Lines.Add ('Coefficients');
  MemoResults.Lines.Add (Format ('%.4f  constant term', [coef [0]]));

  for i := 1 to nTerms - 1 do
    MemoResults.Lines.Add (Format ('%.4f   X^%d', [coef [i], i]));
  MemoResults.Lines.Add (Format ('%.4f  correlation coefficient', [correl]));

  FValid := true;

  paintMe.Refresh;
end;


function point(AX, AY: Integer): TPoint;
begin
  with Result do
  begin
    X := AX;
    Y := AY;
  end;
end;

procedure TFormMain.PlotDiamond(X, Y: Integer);
begin
  paintMe.Canvas.Polygon
    (
      [ point(X,                 Y - diamondSize), 
        point(X + diamondSize,   Y),
        point(X,                 Y + diamondSize),
        point(X - diamondSize,   Y)
      ]
    );
end;

procedure TFormMain.PlotFilledDiamond(X, Y: Integer);
begin
  paintMe.Canvas.Pen.Color := clBlack;
  with paintMe.Canvas.Brush do
  begin
    Style := bsSolid;
    Color := clBlue;
    PlotDiamond(X, Y);
    Style := bsClear;
  end;
end;

//paintMe:TPaintbox code:
//
// Show the "experimental data" as diamonds and
// the curve or line we fitted as a line plot
//
procedure TFormMain.paintMePaint(Sender: TObject);
var
 t:Integer;
 mx,my:Double; // Horizontal and Vertical Pixel Multiplier
 yPixels,xPixels:Double;
 function equation(f_of_x:double):Double;
 var
   order:Integer;
 begin
    result := coef[0];
    for order := 1 to UpDownNumTerms.Position-1 do begin
          result := result + coef[order]*Power(f_of_x,order);// x, x^2, x^3, etc.
    end;
 end;
begin
  if not FValid then exit;
  paintMe.Canvas.Pen.Style := psClear;
  paintMe.Canvas.Brush.Color := paintMe.Color;
  paintMe.Canvas.Brush.Style := bsSolid;
  paintMe.Canvas.Rectangle(0,0,paintMe.Width,PaintMe.Height);
  paintMe.Canvas.Pen.Style := psSolid;


  mx := (paintMe.Width-1) / 20;
  my := (paintMe.Height-1) / 20;
  paintMe.Canvas.Pen.Color := clDkGray;
  for t := 0 to 20 do begin
        yPixels := t*my;
        paintMe.Canvas.MoveTo(0, Round(yPixels));
        paintMe.Canvas.LineTo(paintMe.Width, Round(yPixels));
  end;
  for t := 0 to 20 do begin
        xPixels := t*mx;
        paintMe.Canvas.MoveTo(Round(xPixels),0);
        paintMe.Canvas.LineTo(Round(xPixels),PaintMe.Height);
  end;

  for t := plotStart to plotEnd do begin
    if (x[t]>=0) and (y[t]>=0) then
       PlotFilledDiamond( Round( mx*x[t]), paintMe.Height-Round( my*y[t] ) );
  end;

  yPixels := equation(0{/mx})*my;
  paintMe.Canvas.Pen.Color := clRed;
  paintMe.Canvas.Pen.Style := psSolid;

  paintMe.Canvas.MoveTo(0, paintMe.Height-Round(yPixels));
  for t := 1 to paintMe.Width-1 do begin
      yPixels := equation(t/mx);
      yPixels := yPixels *my;
      paintMe.Canvas.LineTo(  Round( ((t/paintMe.Width)*20) * mx ) , paintMe.Height-Round(yPixels) );
  end;

end;

procedure TFormMain.CheckBoxOriginForceClick(Sender: TObject);
begin
    Calculate_fit (UpDownNumTerms.Position);
end;

end.

